//
// Copyright (c) 1998 Colosseum Builders, Inc.
// All rights reserved.
//
// Colosseum Builders, Inc. makes no warranty, expressed or implied
// with regards to this software. It is provided as is.
//
// Permission to use, redistribute, and copy this file is granted
// without a fee so long as as the following conditions are adhered to:
//
// o The user assumes all risk for using this software. The authors of this
//   software shall be liable for no damages of any kind.
//
// o If the source code is distributed then this copyright notice must
//   remain unaltered and any modification must be noted.
//
// o If this code is shipped in binary format the accompanying documentation
//   should state that "this software is based, in part, on the work of
//   Colosseum Builders, Inc."
//

//
//  Title:  Sample Image Viewer/Format Conversion Application
//
//  Author:  John M. Miano miano@colosseumbuilders.com
//
//  Description:
//
//    MDI Child Class
//
//  August 1, A.D. 2005 - Changed the cut and paste and display functions
//  to use VCL classes.
//
//---------------------------------------------------------------------
#include <vcl.h>

#include <fstream>
#include <cerrno>
#pragma hdrstop

#include "ChildWin.h"
#include "main.h"

//---------------------------------------------------------------------
#pragma resource "*.dfm"

using namespace Colosseum ;

static void ProgressFunction (Colosseum::BitmapImageCoder &coder,
                              void *data,
                              unsigned int currentpass,
                              unsigned int passcount,
                              unsigned int progress,
                              bool &cancel)
{
  unsigned int value = (100 * (currentpass - 1) + progress)
                               / passcount ;
  MainForm->ProgressBar->Position = value ;
  if (progress == 100 && MainForm->ShowProgressive)
  {
    coder.updateImage () ;
    TMDIChild *child = (TMDIChild *) data ;
    child->image_type = Colosseum::JpegImage ;
    child->PaintBox->SetBounds (0, 0, child->image.getWidth (),
                                child->image.getHeight ()) ;
    child->PaintBox->Refresh () ;
  }
  Application->ProcessMessages () ;
  return ;
}

void ImageProgressFunction (Colosseum::BitmapImage &image,
                              void *data,
                              unsigned int currentpass,
                              unsigned int passcount,
                              unsigned int progress,
                              bool &cancel)
{
  MainForm->ProgressBar->Position = (100 * (currentpass - 1) + progress)
                               / passcount ;
  Application->ProcessMessages () ;
  return ;
}

//---------------------------------------------------------------------
__fastcall TMDIChild::TMDIChild(TComponent *Owner)
	: TForm(Owner)
{
  image_type = Colosseum::UnknownImage ;
  gamma_value = 1.0 ;
  is_busy = false ;
  return ;
}
//---------------------------------------------------------------------
void __fastcall TMDIChild::FormClose(TObject *Sender, TCloseAction &Action)
{
  if (! is_busy)
    Action = caFree;
  return ;
}
//---------------------------------------------------------------------
void TMDIChild::ReadImage (const String &filename)
{
  is_busy = true ;

  MainForm->ShowProgressBar () ;
  Cursor = crHourGlass ;
  Caption = filename ;
  image_type = Colosseum::UnknownImage ;

  try
  {
    image_type = Colosseum::ReadImage (
                             filename.c_str (),
                             image,
                             ProgressFunction,
                             reinterpret_cast<void*>(this)) ;
  }
  catch (std::exception &ee)
  {
    Caption = "" ;
    Cursor = crDefault ;
    MainForm->ProgressBar->Visible = false ;
    is_busy = false ;
    throw Exception (ee.what ()) ;
  }
  is_busy = false ;

  copyFromImageToBitmap () ;

  MainForm->ProgressBar->Visible = false ;
  Cursor = crDefault ;
  return ;
}


void __fastcall TMDIChild::PaintBoxPaint(TObject *Sender)
{
  if (image_type == Colosseum::UnknownImage)
    return ;

  PaintBox->Canvas->Draw ( 0, 0, bitmap) ;
  return ;
}
//---------------------------------------------------------------------------


void __fastcall TMDIChild::SetGammaValue (double value)
{
  double adjust = value / gamma_value ;
  Cursor = crHourGlass ;
  Application->ProcessMessages () ;
  copyFromBitmapToImage () ;
  image.GammaCorrect (adjust) ;
  copyFromImageToBitmap () ;
  gamma_value = value ;
  PaintBox->Refresh () ;
  Cursor = crDefault ;
  return ;
}

void TMDIChild::Grayscale ()
{
  Cursor = crHourGlass ;
  Application->ProcessMessages () ;
  copyFromBitmapToImage () ;
  image.ToGrayscale () ;
  copyFromImageToBitmap () ;
  PaintBox->Refresh () ;
  Cursor = crDefault ;
  return ;
}

void TMDIChild::CopyToClipboard ()
{
  bool status = OpenClipboard (Application->Handle) ;
  if (! status)
    throw Exception ("Cannot Open the Clipboard") ;
  status = EmptyClipboard () ;

  Word format ;
  unsigned int handle ;
  HPALETTE palette ;
  bitmap->SaveToClipboardFormat (format, handle, palette) ;

  HANDLE clipboard = SetClipboardData (format, reinterpret_cast<HANDLE>(handle)) ;
  CloseClipboard () ;

  return ;
}

void TMDIChild::CopyFromClipboard ()
{
  OpenClipboard (Application->Handle) ;
  HANDLE hbm = GetClipboardData (CF_BITMAP) ;
  bitmap->LoadFromClipboardFormat (CF_BITMAP, (unsigned int) hbm, 0) ;
  CloseClipboard () ;
  image_type = Colosseum::BmpImage ;

  PaintBox->SetBounds (0, 0, bitmap->Width, bitmap->Height) ;
  return ;
}

void TMDIChild::SetAlpha (unsigned int value)
{
  copyFromBitmapToImage () ;
  BitmapImage::Pixel *pixel = &image [0] ;
  for (unsigned int ii = 0 ; ii < image.getWidth () * image.getHeight () ; ++ ii)
    pixel [ii].alpha = value ;
  copyFromImageToBitmap () ;
  return ;
}

void TMDIChild::MergeImage (const String &filename)
{
  is_busy = true ;

  MainForm->ShowProgressBar () ;
  Cursor = crHourGlass ;
  image_type = Colosseum::UnknownImage ;
  BitmapImage newimage ;
  try
  {
    image_type = Colosseum::ReadImage (
                             filename.c_str (),
                             newimage,
                             ProgressFunction,
                             reinterpret_cast<void*>(this)) ;
  }
  catch (std::exception &ee)
  {
    Caption = "" ;
    Cursor = crDefault ;
    MainForm->ProgressBar->Visible = false ;
    is_busy = false ;
    throw Exception (ee.what ()) ;
  }
  is_busy = false ;

  copyFromBitmapToImage () ;
  image.mergeImage (newimage) ;
  copyFromImageToBitmap () ;

  PaintBox->SetBounds (0, 0, bitmap->Width, bitmap->Height) ;
  MainForm->ProgressBar->Visible = false ;
  PaintBox->Invalidate () ;
  Cursor = crDefault ;
  return ;
}

void __fastcall TMDIChild::FormCreate(TObject *Sender)
{
  bitmap = new Graphics::TBitmap ;
}
//---------------------------------------------------------------------------

void __fastcall TMDIChild::FormDestroy(TObject *Sender)
{
  delete bitmap ; bitmap = 0 ;
}
//---------------------------------------------------------------------------


void TMDIChild::copyFromImageToBitmap ()
{
  bitmap->HandleType = bmDIB ;
  bitmap->PixelFormat = pf32bit ;
  bitmap->Height = image.getHeight () ;
  bitmap->Width = image.getWidth () ;

  for (unsigned int ii = 0 ; ii < image.getHeight () ; ++ ii)
  {
    memcpy (bitmap->ScanLine [ii], &image [ii * image.getWidth ()], 4 * image.getWidth ()) ;
  }


  PaintBox->SetBounds (0, 0, image.getWidth (), image.getHeight ()) ;
  image.clearImage () ;
  return ;
}

void TMDIChild::copyFromBitmapToImage ()
{
  bitmap->HandleType = bmDIB ;
  bitmap->PixelFormat = pf32bit ;
  image.setSize (bitmap->Width, bitmap->Height) ;

  for (unsigned int ii = 0 ; ii < image.getHeight () ; ++ ii)
  {
    memcpy (&image [ii * image.getWidth ()], bitmap->ScanLine [ii], 4 * image.getWidth ()) ;
  }
  return ;
}

