//
// Copyright (c) 2005 Colosseum Builders, Inc.
// All rights reserved.
//
// Colosseum Builders, Inc. makes no warranty, expressed or implied
// with regards to this software. It is provided as is.
//
// See the README.TXT file that came with this software for restrictions
// on the use and redistribution of this file or send E-mail to
// info@colosseumbuilders.com
//

//
// GIF Decoder
//
// Author: John M. Miano miano@colosseumbuilders.com
//

#include <climits>
#include "gifdecoder.h"
#include "gif.h"

using namespace std ;
using namespace Colosseum ;

namespace Colosseum
{

//
//  Description:
//
//    Default Class Constructor
//
GifDecoder::GifDecoder ()
: input_stream (0)
, verbose_flag (false)
{
  return ;
}

//
//  Description:
//
//    Class Destructor
//
GifDecoder::~GifDecoder ()
{
  if (multi_image_mode)
    delete input_stream ;
  input_stream = 0 ;
  return ;
}

//
//  This function reads the GIF header from the input stream and processes it.
//
void GifDecoder::readStreamHeader (GifInputStream &input)
{
  GifHeader header ;
  GifScreenDescriptor screen ;

  // Initialization
  more_images = false ;
  global_color_count = 0 ;
  screen_width = 0 ;
  screen_height = 0 ;

  // Read the GIF Header.
  unsigned int count = input.read (reinterpret_cast<UBYTE1*>(&header), sizeof (header)) ;
  if (count != sizeof (header))
    throw GifError ("GIF: Cannot read GIF header.") ;
  if (verbose_flag)
    header.print (cout) ;

  // The signature should either be GIF87a or GIF89a.
  if (header.file_signature [0] != 'G' || header.file_signature [1] != 'I' || header.file_signature [2] != 'F'
      || header.file_version [0] != '8'  || (header.file_version [1] != '7' && header.file_version [1] != '9')
      || header.file_version [2] != 'a')
      throw GifError ("GIF: Not a GIF stream") ;

  // Read the Screen Descriptor
  count = input.read (reinterpret_cast<UBYTE1*>(&screen), sizeof (screen)) ;
  if (count != sizeof (screen))
    throw GifError ("GIF: Cannot read GIF screen descriptor.") ;

  // 
  screen_width = screen.logical_screen_width ;
  screen_height = screen.logical_screen_height ;

  if (verbose_flag)
    screen.print (cout) ;

  if (screen.globalColorTableFlag ())
  {
    global_color_count = 1 << screen.globalColorTableSize () + 1 ;
    for (unsigned int ii = 0 ; ii < global_color_count ; ++ ii)
    {
      count = input.read (reinterpret_cast<UBYTE1*>(&global_color_table [ii]), sizeof (global_color_table [ii])) ;
      if (count != sizeof (global_color_table [ii]))
        throw GifError ("GIF: Error reading global color table in GIF file.") ;
    }
  }
  else
  {
    global_color_count = 0 ;
  }
  return ;
}

//
//  This function reads until the next image within the GIF stream.
//
void GifDecoder::readToNextImage (GifInputStream &input)
{
  unsigned int blocktype ;
  unsigned int controllabel ;
  unsigned int size ;
  char buffer [256] ;

  GifGraphicControlBlock graphiccontrolbblock ;
  GifApplicationBlock applicationblock ;
  GifPlainTextBlock textblock ;

  while (input.moreData ())
  {
    blocktype = input.getByte () ;
    switch (blocktype)
    {
    case GifTrailer:
      if (multi_image_mode)
        delete input_stream ;
      input_stream = 0 ;
      more_images = false ;
      return ;
    case GifImageSeparator:
      more_images = true ;
      return ;
    case GifExtension:
      controllabel = input.getByte () ;
      switch (controllabel)
      {
      case GifPlainTextExtension:
        size = input.read (reinterpret_cast<UBYTE1*>(&textblock), sizeof (textblock)) ;
        if (size != sizeof (textblock))
          throw GifError ("GIF: Cannot read Plain Text Extension Block") ;
        if (textblock.block_size != (sizeof (textblock) - sizeof (textblock.block_size)))
              throw GifError ("GIF: Corrupt GIF file - wrong size in GIF extension") ;
        if (verbose_flag)
          textblock.print (cout) ;
        size = input.getByte () ;
        while (size != 0 && input.moreData ())
        {
          input.read (buffer, size) ;
          size = input.getByte () ;
        }
        if (size != 0)
          throw GifError ("GIF: Cannot read Plaint Text data.") ;
        break ;
      case GifGraphicControlExtension:
        input.read (reinterpret_cast<UBYTE1*>(&graphiccontrolbblock), sizeof (graphiccontrolbblock)) ;
        if (graphiccontrolbblock.block_size != (sizeof (graphiccontrolbblock) - sizeof (graphiccontrolbblock.block_size)))
          throw GifError ("GIF: Corrupt GIF file - wrong size in Graphic Control Block") ;
        if (verbose_flag)
          graphiccontrolbblock.print (cout) ;
        blocktype = input.getByte () ;
        if (blocktype != 0)
          throw GifError ("GIF: Corrupt GIF file - missing terminator") ;

        transparent_flag = graphiccontrolbblock.transparentColorFlag () ;
        transparent_color = graphiccontrolbblock.transparent_color ;
        break ;
      case GifCommentExtension:
        size = input.getByte () ;
        input.read (buffer, size) ;
        if (verbose_flag)
        {
          cout << "{ Comment Extension" << endl ; ;
          cout << buffer << endl ;
          cout << "}" << endl ; ;
        }
        break ;
      case GifApplicationExtension:
        input.read (reinterpret_cast<UBYTE1*>(&applicationblock), sizeof (applicationblock)) ;
        if (applicationblock.block_size != 11)
          throw GifError ("GIF: Corrupt GIF file - invalid Application Extension Length") ;
        if (verbose_flag)
          applicationblock.print (cout) ;

        size = input.getByte () ;
        while (size != 0 && input.moreData ())
        {
          input.read (buffer, size) ;
          size = input.getByte () ;
        }
        if (size != 0)
          throw GifError ("GIF: Cannot read Application Extension data") ;
        break ;
      default:
        throw GifError ("GIF: Invalid GIF Extension") ;
      }
      break ;
    default:
      throw GifError ("GIF: Invalid GIF block") ;
    }
  }
  // Reached the end of stream without the GifTrailer marker.
  throw GifError ("GIF: Missing GIF Trailer marker.") ;
} // ReadToNextImage

//
//  This function sets the multi_image_mode property. This function determines
//  whether the decode allows processing multiple images from the input stream.
//
void GifDecoder::setMultiImageMode (bool state)
{
  if (input_stream != 0)
    throw GifError ("GIF: Cannot change multi-image mode with active input stream.") ;

  multi_image_mode = state ;
  return ;
}

//
//  This function reads the next image from the input stream. It assumes
//  the read position is at the start of an image.
//
void GifDecoder::readImage (GifInputStream &input, BitmapImage &image)
{
  if  (input_stream != 0)
    throw GifError ("GIF: Input stream already open.") ;

  readStreamHeader (input) ;
  input_stream = &input ;
  readToNextImage (input) ;
  if (! more_images)
    throw GifError ("GIF: No image in input stream.") ;

  readNextImage (image) ;
  if (! (more_images && multi_image_mode))
    input_stream = 0 ;

  return ;
}

//
//  This function writes the next value to the image.
//
void GifDecoder::outputToImage (UBYTE1 data,
                                BitmapImage &image)
{
  // All of these constants are used to support interlaced files.
  const unsigned int MAXPASSES = 4 ;
  // We have one extra entry here for debugging. If
  const unsigned int passstart [MAXPASSES + 1] = { 0, 4, 2, 1, 0x7FFFFFFF} ;
  const unsigned int passcount [MAXPASSES] =  {8, 8, 4, 2} ;
  const unsigned int passrows [MAXPASSES] = {8, 4, 2, 1} ;

  unsigned int offset ;
  unsigned int limit ;

  if (row_ptr >= image.getHeight ())
    throw GifError ("GIF: Corrupt input stream. Compressed data past image.") ;

  if (is_interlaced)
  {
    limit = image.getWidth () * image.getHeight () ; // Mark the end of the image.
    offset = row_ptr * image.getWidth () + col_ptr ; // Locate the current pixel.

    for (unsigned int ii = 0 ; ii < passrows [interlace_pass] && offset < limit ; ++ii)
    {
      // Here we fill in the pixel and all the pixels between the
      // interlaced rows.
      image [offset].red   = color_table [data].red ;
      image [offset].green = color_table [data].green ;
      image [offset].blue  = color_table [data].blue ;
      if (transparent_flag && data == transparent_color)
        image [offset].alpha = 0x00 ;
      else
        image [offset].alpha = 0xFF ;
      offset += image.getWidth () ; // Advance to the next row.
    }
    ++ col_ptr ;
    if (col_ptr == image.getWidth ())
    {
      col_ptr = 0 ;
      row_ptr += passcount [interlace_pass] ;
      callProgressFunction (interlace_pass + 1,
                            MAXPASSES, (100 * row_ptr) / image.getHeight ()) ;
      if (row_ptr >= image.getHeight ())
      {
        ++ interlace_pass ;

        // Handle something stupid like a one-pixel high interlaced image.
        while (passstart [interlace_pass] >= image.getHeight ()
               && interlace_pass < MAXPASSES)
        {
          ++ interlace_pass ;
        }
        row_ptr = passstart [interlace_pass] ;
      }
    }
  }
  else
  {  // Noninterlaced file.
    offset = row_ptr * image.getWidth () + col_ptr ;
    image [offset].red   = color_table [data].red ;
    image [offset].green = color_table [data].green ;
    image [offset].blue  = color_table [data].blue ;
    if (transparent_flag && data == transparent_color)
      image [offset].alpha = 0x00 ;
    else
      image [offset].alpha = 0xFF ;

    ++ col_ptr ;
    if (col_ptr == image.getWidth ())
    {
      ++ row_ptr ;
      callProgressFunction (1, 1, (100 * row_ptr) / image.getHeight ()) ;
      col_ptr = 0 ;
    }
  }
  return ;
}


//
//  This function converts a code within the input stream into a string
//  of characters and writes the characters to the output images.
//
//  Parameters:
//      code : The code to expand
//      image : The output image.
//
void GifDecoder::outputCode (unsigned int incode,
                             BitmapImage &image)
{
  int sp ;
  UBYTE1 stack [(1 << GifMaxBitsPerCode) + 1] ;
  unsigned int lastcode ;
  unsigned int code = incode ;

  // Push the values on the stack in reverse order.
  sp = 0 ;
  do
  {
    ++ sp ;
    if (sp >= (1 << GifMaxBitsPerCode) + 1)
      throw GifError ("GIF: Corrupt compressed data.") ;
    if (code >= (1 << GifMaxBitsPerCode))
      throw GifError ("GIF: Corrupt compressed data.") ;
    stack [sp] = lzw_dictionary [code].data ;
    lastcode = code ;
    code = lzw_dictionary [code].parent ;
  } while (lastcode >= clear_code) ;

  first_character = stack [sp] ;

  // Pop the values form the stack in the correct order.
  while (sp > 0)
  {
    outputToImage (stack [sp], image) ;
    -- sp ;
  }
}
//
//  Description:
//    This function reads the next image from the GIF stream.
//
// Parameters:
//    image: The image to read
//
void GifDecoder::readNextImage (BitmapImage &image)
{
  GifImageDescriptor id ;
  unsigned int colorcount ;

  if (! more_images)
    throw GifError ("GIF: No more images.") ;

  if (input_stream == 0)
     throw GifError ("GIF: No input stream") ;

  if (! more_images)
    throw GifError ("GIF: No more images in input stream.") ;

  if (input_stream->read (reinterpret_cast<UBYTE1*>(&id), sizeof (id)) != sizeof (id))
    throw GifError ("GIF: Error reading GIF image descriptor") ;
  if (verbose_flag)
    id.print () ;

  // Save the image position so the user can retrieve it later to position
  // the image on the logical screen.
  image_position_top = id.top_position ;
  image_position_left = id.left_position ;

  // Interlaced handling.
  is_interlaced = id.interlaceFlag () ;
  interlace_pass = 0 ;

  // Determine if this image uses the local color table.
  if (id.localColorTableFlag () &&  id.colorTableSize () != 0)
  {
    // Use a local color table.
    colorcount = 1 << (id.colorTableSize () + 1) ;
    unsigned int count = input_stream->read (reinterpret_cast<UBYTE1*>(color_table), sizeof (color_table [0]) * colorcount) ;
    if (count != sizeof (color_table [0]) * colorcount)
      throw GifError ("GIF: Error reading local color table in GIF file.") ;
  }
  else
  {
    // No local color table.

    // If there is no global color table as well then we're in trouble.
    if (global_color_count == 0)
      throw GifError ("GIF: Image contains no color table") ;

    colorcount = global_color_count ;
    for (unsigned int ii = 0 ; ii < colorcount ; ++ ii)
      color_table [ii] = global_color_table [ii] ;
  }

  // Allocate the image buffer.
  image.setSize (id.image_width, id.image_height) ;

  // Initialize the output position in the image.
  row_ptr = 0 ;
  col_ptr = 0 ;

  // The minimum code size comes before the compressed data.
  min_code_size = input_stream->getByte () ;

  // Read the comrpessed data.
  initialize () ;
  input_stream->enterBitMode ();
  decompress (image) ;
  input_stream->exitBitMode () ;

  // Get in a position where the user can read the next image in the stream.
  if (multi_image_mode)
    readToNextImage (*input_stream) ;

  return ;
}

//
// Description:
//    This function calls the progress function.
//
void GifDecoder::callProgressFunction (unsigned int pass, unsigned int passcount, unsigned int progress)
{
  bool abort ;
  unsigned int percent ;

  if (progress_function == 0)
    return ;

  abort = false ;
  if (progress > 100)
    percent = 100 ;
  else
    percent = progress ;

  progress_function (*this,
                     progress_data,
                     pass,
                     passcount,
                     percent,
                     abort) ;
  if (abort)
    throw GraphicsAbort () ;

  return ;
}


//
//  This function handles clear codes in the input stream.
//
void GifDecoder::initialize ()
{
  // Initialize the LZW Dictionary
  for (unsigned int ii = 0 ; ii < (1 << min_code_size) ; ++ ii)
  {
    lzw_dictionary [ii].data = ii ;
    lzw_dictionary [ii].parent = 0 ;
  }
  for (unsigned int ii = 1 << min_code_size ;
       ii < sizeof (lzw_dictionary) / sizeof (lzw_dictionary [0]) ;
       ++ ii)
  {
    // Load with junk to aid debugging.
    lzw_dictionary [ii].data = 0xFF ;
    lzw_dictionary [ii].parent = 0xFFFF ;
  }

  // Generate values based upon the code size
  code_size  = min_code_size + 1 ;  // Initial code size
  clear_code = 1 << min_code_size ;  // Reset decompressor code
  end_code   = clear_code + 1 ; // End of compressed stream code
  next_code  = end_code + 1 ;  // Next free code
  return ;
}


//
//  This function does the LZW decompression of the compressed data.
//
void GifDecoder::decompress (BitmapImage &image)
{
  unsigned int code ;
  unsigned int lastcode ;
  // Get the first code in the stream. A clear code can come anywhere and
  // some compressors like to put them at the start of the stream.
  code = input_stream->nextCode (code_size) ;
  while (code == clear_code)
    code = input_stream->nextCode (code_size) ;
  outputCode (code, image) ;

  while (true)
  {
    lastcode = code ;
    // See if we need to increase the code size.
    if (next_code >= (1 << code_size) && code_size < GifMaxBitsPerCode)
      ++ code_size ;

    code = input_stream->nextCode (code_size) ;
    if (code == end_code)
    {
      return ; // End Code => All Done
    }
    else if (code == clear_code)
    {
      // Clearcode => Reset the compressor and start over.
      initialize () ;
      code = input_stream->nextCode (code_size) ;
      while (code == clear_code)
        code = input_stream->nextCode (code_size) ;

      if (code == end_code)
          return ;

      outputCode (code, image) ;
    }
    else if (code < next_code)
    {
      // A code that has already been defined.
      outputCode (code, image) ;
      lzw_dictionary [next_code].parent = lastcode ;
      lzw_dictionary [next_code].data = first_character ;
      ++ next_code ;
    }
    else
    {
      if (code != next_code)
        throw GifError ("GIF: Corrupt compressed data.") ;
      // Special Case of undefined code.
      lzw_dictionary [next_code].parent = lastcode ;
      lzw_dictionary [next_code].data = first_character ;
      ++ next_code ;
      outputCode (code, image) ;
    }
  }
}


//
//  This function reads the first image from a GIF file.
//
//  Parameters:
//    filename : the input file name
//    image : the image to write to.
//
void GifDecoder::readImageFile (const std::string &filename, BitmapImage &image)
{
  if (input_stream != 0)
    throw GifError ("GIF: Already have open input stream") ;

  GifInputFileStream *input = new GifInputFileStream ;
  try
  {
    input->open (filename) ;
    readImage (*input, image) ;
  }
  catch (...)
  {
    delete input_stream ;
    throw ;
  }

  if (! multi_image_mode && input_stream != 0)
  {
    if (input_stream != input)
      throw GifError ("GIF: Internal Error - Corrupt input stream.") ;

    closeInputStream () ;
  }
  return ;
}

void GifDecoder::closeInputStream ()
{
  if (input_stream == 0)
    return ;

  delete input_stream ;
  input_stream = 0 ;
  return ;
}

bool GifDecoder::activeInputStream ()
{
  if (input_stream == 0)
    return false ;
  else
    return true ;
}


}
